import java.io.*;
import java.net.*;

public class DebugUDPThread implements Runnable {

	private Thread sendThread = null;
	private Thread receiveThread = null;
	private ReceiveData receiveData = null;
	private DebugUDPUI ui = null;
	private DebugUDPState st;
	private DatagramSocket dgSocket = null;
	private DatagramPacket dgPacketSend = null;
	private byte[] outputData;

	public DebugUDPThread(DebugUDPUI userInterface, DebugUDPState state) {
		ui = userInterface;
		st = state;
	}

	public void startRunning() {
		sendThread = new Thread(this,"SendThread");
		sendThread.start();
	}

	public void stopRunning() {
		st.hasPeer = false;
		st.isRunning = false;
		dgSocket.close();
		synchronized (this) {
			notify();
		}
	}

	public void run() {
		try {
			//st.hasPeer = false;  // ensure consistency
			if (st.localPort == 0) {
				// Bind to any available port
				dgSocket = new DatagramSocket(); 
				st.localPort = dgSocket.getLocalPort();
			} else {
				// Bind to a specific port
				dgSocket = new DatagramSocket(st.localPort); 
			}
			if (st.remotePort > 0 && st.remotePort < 65536) {
				if (!st.remoteHostStr.equals("")) {
					st.hasPeer = true;
				}
			}
			ui.updateState();
			startReceiverThread();
			waitForOutputData();
		} catch (SocketException ex) {
			ui.popupErrorMessage(ex.getMessage());
		} catch (IllegalArgumentException ex) {
			ui.popupErrorMessage(ex.getMessage());
		} finally {
			st.hasPeer = false;    // clean up in case there was an exception
			st.isRunning = false;
			ui.updateState();
		}
	}

	private void startReceiverThread() {
		receiveData = new ReceiveData();
		receiveThread = new Thread(receiveData,"ReceiveThread");
		receiveThread.start();
	}

	private synchronized void waitForOutputData() {
		while (st.isRunning) {
			try {
				wait();
				if (st.hasPeer) {
					// a datagram packet must be sent
					if (st.remoteHostStr.equals("")) {
						UnknownHostException ex = new UnknownHostException();
						throw (ex);
					};
					InetAddress address = InetAddress.getByName(st.remoteHostStr);
					dgPacketSend = new DatagramPacket(outputData,
							outputData.length,
							address,
							st.remotePort);
					//System.out.println("sending data");
					dgSocket.send(dgPacketSend);
				} // else the user clicked stop, but we probably don't care here
			} catch (InterruptedException ex) {
				ui.popupErrorMessage("Interrupted: " + ex.getMessage());
			} catch (UnknownHostException ex) {
				ui.popupErrorMessage("Unkown remote host");
			} catch (IOException ex) {
				ui.popupErrorMessage(ex.getMessage());
			}
		} // while (st.isRunning)
	}

	public synchronized void sendMessage(byte[] buf, int bufLength) {
		outputData = new byte[bufLength];
		System.arraycopy(buf,0,outputData,0,bufLength);
		notify();
	}

	// Class for running the receiver thread
	public class ReceiveData implements Runnable {
		private byte[] buf;
		private DatagramPacket dgPacketReceive;
		public ReceiveData() {
			buf = new byte[st.packetLength];
			dgPacketReceive = new DatagramPacket(buf, st.packetLength);
		}
		public void run() {
			//System.out.println("receive thread running");
			while (st.isRunning) {
				try {
					dgSocket.receive(dgPacketReceive);
					//System.out.println("packet received");
					st.remoteHostStr = dgPacketReceive.getAddress().getHostAddress();
					st.remotePort = dgPacketReceive.getPort();
					st.hasPeer = true;
					// display received data
					ui.displayReceivedData(dgPacketReceive.getData(),
							dgPacketReceive.getLength());
					//String newstr = new String(buf);
					//System.out.println(newstr);
					ui.updateState(); // may not be necessary every time
				} catch (IOException ex) {
					// maybe we want to be here because the user clicked stop?
					if (st.isRunning) { 
						ui.popupErrorMessage(ex.getMessage());
					}
				}
			}
		}
	}

}
