import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.reflect.InvocationTargetException;

public class DebugUDPTool implements ActionListener, CaretListener, FocusListener, DebugUDPUI {

	// State information
	DebugUDPState st;

	// frames and panels
	private JFrame mainFrame;
	private JPanel mainPanel;
	private JPanel configAndEntryPanel;
	private JPanel configPanel;
	private JPanel sendPanel;
	private JPanel textDisplayPanel;
	private JPanel hexDisplayPanel;
	private JPanel statusRowPanel;

	// Configuration
	private JLabel localPortEntryLabel;
	private JTextField localPortEntry;
	private JLabel remoteHostEntryLabel;
	private JTextField remoteHostEntry;
	private JLabel remotePortEntryLabel;
	private JTextField remotePortEntry;
	private JButton startButton;
	private JButton stopButton;
	private JLabel msgTerminatorEntryLabel;
	private JTextField msgTerminatorEntry;

	// Transmit
	private final static int numSendRows = 5;
	private JPanel[] sendRow;
	private JTextField[] sendTextEntry;
	private JButton[] sendTextButton;

	// Receive
	private boolean textDisplayAreaHasFocus;
	private JTextArea textDisplayArea;
	private JScrollPane textDisplayPane;
	private JTextArea hexDisplayArea;
	private JScrollPane hexDisplayPane;
	private Font displayFont;

	// Status row
	private int selectionStartInt;
	private int selectionEndInt;
	private JLabel statusLabel;
	private JButton swapSelection;
	private JButton clearButton;
	private JCheckBox autoClear; 

	private DebugUDPThread netThread;

	// Miscellaneous
	private String errorMessage;

	// Constructor
	public DebugUDPTool() {
		st = new DebugUDPState();
		netThread = new DebugUDPThread(this, st);
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				createAndShowGUI();
			}
		} );
	}
	
	// Create and show GUI
	private void createAndShowGUI() {
		// Create frame and panel
		JFrame.setDefaultLookAndFeelDecorated(true);
		mainFrame = new JFrame("UDP Debugging Tool by Michael Siegenthaler");
		mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel,BoxLayout.PAGE_AXIS));
		configAndEntryPanel = new JPanel();
		configAndEntryPanel.setLayout(new GridLayout(1,0));
		configPanel = new JPanel();
		configPanel.setLayout(new GridLayout(0,2));
		configPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Configuration"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		sendPanel = new JPanel();
		sendPanel.setLayout(new GridLayout(0,1));
		sendPanel.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createTitledBorder("Transmit"),
					BorderFactory.createEmptyBorder(5,5,5,5)));
		textDisplayPanel = new JPanel();
		textDisplayPanel.setLayout(new BorderLayout());
		textDisplayPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Received Text"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		hexDisplayPanel = new JPanel();
		hexDisplayPanel.setLayout(new BorderLayout());
		hexDisplayPanel.setBorder(BorderFactory.createCompoundBorder(
				       BorderFactory.createTitledBorder("Received Data"),
				       BorderFactory.createEmptyBorder(5,5,5,5)));
		statusRowPanel = new JPanel();
		statusRowPanel.setLayout( new GridLayout(1,0) );
		statusRowPanel.setBorder( BorderFactory.createEmptyBorder(5,5,5,5) );

		// Configuration
		localPortEntryLabel = new JLabel("Local port:");
		localPortEntry = new JTextField(15);
		//localPortEntry.setText("2400");
		remoteHostEntryLabel = new JLabel("Remote host:");
		remoteHostEntry = new JTextField(15);
		remoteHostEntry.setText("127.0.0.1");
		remoteHostEntry.addFocusListener(this);
		remotePortEntryLabel = new JLabel("Remote port:");
		remotePortEntry = new JTextField(15);
		//remotePortEntry.setText("2400");
		remotePortEntry.addFocusListener(this);
		startButton = new JButton("Start");
		startButton.addActionListener(this);
		startButton.setMnemonic('s');
		stopButton = new JButton("Stop");
		stopButton.addActionListener(this);
		stopButton.setMnemonic('o');
		msgTerminatorEntryLabel = new JLabel("Message terminator:");
		msgTerminatorEntry = new JTextField(15);
		msgTerminatorEntry.setText("\\r\\n");

		configPanel.add(localPortEntryLabel);
		configPanel.add(localPortEntry);
		configPanel.add(remoteHostEntryLabel);
		configPanel.add(remoteHostEntry);
		configPanel.add(remotePortEntryLabel);
		configPanel.add(remotePortEntry);
		configPanel.add(startButton);
		configPanel.add(stopButton);
		configPanel.add(msgTerminatorEntryLabel);
		configPanel.add(msgTerminatorEntry);

		// Transmit
		sendRow = new JPanel[numSendRows];
		sendTextEntry = new JTextField[numSendRows];
		sendTextButton = new JButton[numSendRows];
		for ( int i = 0 ; i < numSendRows ; i++) {
			sendRow[i] = new JPanel();
			sendRow[i].setLayout(new BoxLayout(sendRow[i],BoxLayout.LINE_AXIS));

			sendTextEntry[i] = new JTextField(20);
			sendTextEntry[i].addActionListener(this);

			sendTextButton[i] = new JButton("Send");
			sendTextButton[i].addActionListener(this);

			sendRow[i].add(sendTextEntry[i]);
			sendRow[i].add(sendTextButton[i]);

			sendPanel.add(sendRow[i]);
		}

		// Receive
		displayFont = new Font("Monospaced", Font.PLAIN, 12);
		textDisplayArea = new JTextArea(1,1);
		textDisplayArea.setFont(displayFont);
		textDisplayArea.setEditable(false);
		textDisplayArea.addFocusListener(this);
		textDisplayArea.addCaretListener(this);
		textDisplayPane = new JScrollPane(textDisplayArea);
		textDisplayPane.setPreferredSize(new Dimension(600,150));

		hexDisplayArea = new JTextArea(1,1);
		hexDisplayArea.setFont(displayFont);
		hexDisplayArea.setEditable(false);
		hexDisplayArea.setLineWrap(true);
		hexDisplayArea.setWrapStyleWord(true);
		hexDisplayArea.addFocusListener(this);
		hexDisplayArea.addCaretListener(this);
		hexDisplayPane = new JScrollPane(hexDisplayArea);
		hexDisplayPane.setPreferredSize(new Dimension(600,150));

		textDisplayPanel.add(textDisplayPane);
		hexDisplayPanel.add(hexDisplayPane);

		// Status row
		clearButton = new JButton("Clear");
		clearButton.addActionListener(this);
		clearButton.setMnemonic('r');

		swapSelection = new JButton("Swap");
		swapSelection.addActionListener(this);
		swapSelection.setMnemonic('w');

		autoClear = new JCheckBox("Display one packet");
		//autoClear.setSelected(true);
		autoClear.setMnemonic('d');

		statusLabel = new JLabel("");
		updateStatusLabel(true);

		statusRowPanel.add(statusLabel);
		statusRowPanel.add(swapSelection);
		statusRowPanel.add(clearButton);
		statusRowPanel.add(autoClear);

		// Set up panel hierarchy
		configAndEntryPanel.add(configPanel);
		configAndEntryPanel.add(sendPanel);
		mainPanel.add(configAndEntryPanel);
		mainPanel.add(textDisplayPanel);
		mainPanel.add(hexDisplayPanel);
		mainPanel.add(statusRowPanel);

		// Put everything in a nice, consistent state
		updateGeneralState();

		// Show the gui
		mainFrame.getContentPane().add(mainPanel,BorderLayout.CENTER);
		mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		mainFrame.pack();
		mainFrame.setLocationRelativeTo(null);
		mainFrame.setVisible(true);

	}

	// Event handler
	public void actionPerformed(ActionEvent e) {

		Object src = e.getSource();

		if(src == startButton) {
			try {
				st.localPort = Integer.parseInt(localPortEntry.getText());
			} catch (NumberFormatException ex) {
				st.localPort = 0;
			}
			try {
				st.remotePort = Integer.parseInt(remotePortEntry.getText());
			} catch (NumberFormatException ex) {
				st.remotePort = 0;
			}
			st.isRunning = true;
			st.remoteHostStr = remoteHostEntry.getText();
			st.localPortStr = localPortEntry.getText();
			//st.remotePortStr = remotePortEntry.getText();
			netThread.startRunning();
		} else if (src == stopButton) {
			netThread.stopRunning();
		} else if (src == clearButton) {
			// Clear displays
			st.textReceived = "";
			st.hexReceived = "";
			textDisplayArea.setText("");
			hexDisplayArea.setText("");
			st.rxLength = 0;
			st.txLength = 0;
			updateStatusLabel(true);
		} else if (src == swapSelection) {
			if (textDisplayAreaHasFocus) {
				hexDisplayArea.requestFocus();
				hexDisplayArea.select(selectionStartInt*3,selectionEndInt*3-1);
			} else {
				textDisplayArea.requestFocus();
				textDisplayArea.select(selectionStartInt,selectionEndInt);
			}
			updateStatusLabel(false);
		} else {
			for ( int i = 0 ; i < numSendRows ; i++) {
				if (src == sendTextEntry[i] || src == sendTextButton[i]) {
					// Send text
					ConvertIOData converter = new ConvertIOData(); 
					int bufLength;
					String str = sendTextEntry[i].getText()
						+ msgTerminatorEntry.getText();
					byte[] buf = str.getBytes();
					buf = converter.binarize(buf);
					bufLength = converter.lastBinarizeLength();
					if (autoClear.isSelected()) {
						st.txLength = bufLength;
					} else {
						st.txLength += bufLength;
					}
					netThread.sendMessage(buf,bufLength);
					updateStatusLabel(true);
				}
			}
		}

	}

	public void caretUpdate(CaretEvent e) {
		int start = e.getDot();
		int end = e.getMark();
		if ( e.getSource() == textDisplayArea ) {
			textDisplayAreaHasFocus = true;
		} else if ( e.getSource() == hexDisplayArea ) {
			start++;
			start /= 3;
			end++;
			end /= 3;
			textDisplayAreaHasFocus = false;
		}
		if(start < end) {
			selectionStartInt = start;
			selectionEndInt = end;
		} else {
			selectionStartInt = end;
			selectionEndInt = start;
		}
		if (st.displayAreaHasFocus) {
			SwingUtilities.invokeLater( new Runnable() {
				public void run() {
					updateStatusLabel();
				}
			} );
		}
	}

	public void focusGained(FocusEvent fe) {
		Object src = fe.getSource();
		if (src == textDisplayArea) {
			st.displayAreaHasFocus = true;
			updateStatusLabel();
		} else if (src == hexDisplayArea) {
			st.displayAreaHasFocus = true;
			updateStatusLabel();
		}
	}

	public void focusLost(FocusEvent fe) {
		Object src = fe.getSource();
		boolean hadPeerBefore = st.hasPeer;
		if (src == textDisplayArea) {
			if (!hexDisplayArea.isFocusOwner()) {
				st.displayAreaHasFocus = false;
				updateStatusLabel();
			}
		} else if (src == hexDisplayArea) {
			if (!textDisplayArea.isFocusOwner()) {
				st.displayAreaHasFocus = false;
				updateStatusLabel();
			}
		} else if (src == remoteHostEntry || src == remotePortEntry) {
			if (st.isRunning) {
				if (remoteHostEntry.getText().equals("")) {
					// remote host is blank; thus there is no known peer
					st.hasPeer = false;
				} else {
					// remote host contains some string
					st.remoteHostStr = remoteHostEntry.getText(); 
					try {
						st.remotePort = Integer.parseInt(remotePortEntry.getText());
						if (st.remotePort > 0 && st.remotePort < 65536) { 
						// remote port appears valid
							st.hasPeer = true;
						} else {
						// remote port is out of range
							st.hasPeer = false;
						}
					} catch (NumberFormatException ex) {
						// remote port appears invalid
						st.hasPeer = false;
					}
				}
				if (hadPeerBefore != st.hasPeer) {
					//netThread.notifyFromOtherClass();
					updateGeneralState();
				}
			}
		}
	}

	// Safe method for calling updateGeneralState() from a non-event-dispatching thread
	public void updateState() {
		try {
			SwingUtilities.invokeAndWait( new Runnable() {
				public void run() {
					updateGeneralState();
				}
			} );
		} catch (InterruptedException ex) {
		} catch (InvocationTargetException ex) {
		}
	}

	// Show an error message
	public void popupErrorMessage(String message) {
		errorMessage = message;
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(mainFrame,errorMessage,
					"Error",JOptionPane.ERROR_MESSAGE);
			}
		} );
	}	

	// Receive some data and display it
	public void displayReceivedData(byte[] data, int dataLength) {
		ConvertIOData converter = new ConvertIOData(true);
		converter.convert(data,dataLength);
		if (autoClear.isSelected()) {
			st.rxLength = dataLength;
			st.textReceived = converter.getText();
			st.hexReceived = converter.getHex();
		} else {
			st.rxLength += dataLength;
			st.textReceived += converter.getText() + "\n";
			st.hexReceived += converter.getHex() + "  \n";
		}
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				textDisplayArea.setText(st.textReceived);
				hexDisplayArea.setText(st.hexReceived);
				updateStatusLabel(true);
			}
		} );
	}

	/* 
	// Display the remote address and port
	public void displayRemoteAddressAndPort(String remoteAddress, int remotePort) {
		st.remoteHost = remoteAddress;
		st.remotePort = String.valueOf(remotePort);
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				clientIPDisplay.setText(st.remoteHost);
				clientPortDisplay.setText(st.remotePort);
			}
		} );
	}
	*/

	// Change UI to reflect changes in the connection state
	private void updateGeneralState() {
		if (st.isRunning) {
			// network thread is running
			startButton.setEnabled(false);
			stopButton.setEnabled(true);
			localPortEntry.setEnabled(false);
			localPortEntry.setText(Integer.toString(st.localPort));
			if (st.hasPeer) {
				remotePortEntry.setText(Integer.toString(st.remotePort));
				remoteHostEntry.setText(st.remoteHostStr);
			}
		} else {
			// network thread is not running
			startButton.setEnabled(true);
			stopButton.setEnabled(false);
			localPortEntry.setEnabled(true);
			localPortEntry.setText(st.localPortStr);
			//remotePortEntry.setText(st.remotePortStr);
		}
		updateSendFieldState();
		updateStatusLabel();
	}

	// Enable or disable send buttons based on whether a connection exists or not
	private void updateSendFieldState() {
		for ( int i = 0 ; i < numSendRows ; i++ ) {
			if (st.hasPeer) {
				sendTextButton[i].setEnabled(true);
			} else {
				sendTextButton[i].setEnabled(false);
			}
		}
	}

	private void updateStatusLabel() {
		if (st.displayAreaHasFocus) {
			updateStatusLabel(false);
		} else {
			updateStatusLabel(true);
		}
	}

	// Update the status label
	private void updateStatusLabel(boolean showRxTx) {
		//if (st.isAttemptingToConnect) {
		//	statusLabel.setText("Connecting...");
		//} else {
			if (showRxTx) {
				statusLabel.setText("Rx: " + st.rxLength + ", Tx: " + st.txLength);
			} else {
				statusLabel.setText("Selected: " + selectionStartInt + ", " + selectionEndInt);
			}
		//}
	}

	// Main program
	public static void main(String[] args) {
		new DebugUDPTool();
	}

}
